#ifndef _CMEMPOOL_CPP
#define _CMEMPOOL_CPP
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <Windows.H>
#include <Stdio.H>
#include <Stdlib.H>

#include "CMemPool.H"
#include "../../SharedSource/Common.H"

#ifdef _SQLESERVER
	#include "../../SQLEServer/Source/Routines.h"
#elif _SQLECLIENT
	#include "../../SQLEClient/Source/Routines.h"
#else
	#error "You must define either _SQLESERVER or _SQLECLIENT"
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CMemoryPool gMem;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int CMemoryPool::GetMemAllocationCount(void)
{
	return iUserAllocs;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int CMemoryPool::GetSlotAllocationCount(void)
{
	return UserAlloc.iAllocated;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CMemoryPool::FreeAll(void)
{
	int iSlot = 0;

	//Free all the memory stored in the AllocPool structure.
	while(iSlot < UserAlloc.iAllocated)
	{
		if(UserAlloc.AllocPool[iSlot].bFree == false)
		{
			Free((void *)UserAlloc.AllocPool[iSlot].iAddress);
		}
		iSlot++;
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int CMemoryPool::GetSlotByAddress(int iAddress)
{
	EnterCriticalSection(&csMemLock);

	int iSlot = 0;

	while(iSlot < UserAlloc.iAllocated)
	{
		if(UserAlloc.AllocPool[iSlot].bFree == false)
		{
			if(UserAlloc.AllocPool[iSlot].iAddress == iAddress)
			{
				LeaveCriticalSection(&csMemLock);
				return iSlot;
			}
		}
		iSlot++;
	}

	char sText[1024];
	sprintf(sText, "Error in GetSlotByAddress: Application attempted to access an unmanaged memory address at: %d.", iAddress);
	WriteAnEvent(sText, EVENT_ERROR);

	#ifdef _DEBUG_MEMPOOL
	__debugbreak();
	#endif

	LeaveCriticalSection(&csMemLock);

	return -1; //Did not find the address specified.
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int CMemoryPool::FindFreeAllocSlot(void)
{
	EnterCriticalSection(&csMemLock);

	int iSlot = 0;

	while(iSlot < UserAlloc.iAllocated)
	{
		if(UserAlloc.AllocPool[iSlot].bFree)
		{
			LeaveCriticalSection(&csMemLock);
			return iSlot;
		}
		iSlot++;
	}

	int iMemory = (sizeof(ALLOCPOOL) * UserAlloc.iAllocated) + (sizeof(ALLOCPOOL) * UserAlloc.iAllocIncrement);
	if((UserAlloc.AllocPool = (ALLOCPOOL *) realloc(UserAlloc.AllocPool, iMemory)) == NULL)
	{
		char sText[1024];
		sprintf(sText, "Error in FindFreeAllocSlot: Failed to reallocate the managed AllocPool structure for %d bytes.", iMemory);
		WriteAnEvent(sText, EVENT_ERROR);

		#ifdef _DEBUG_MEMPOOL
		__debugbreak();
		#endif

		return -1;
	}

	UserAlloc.iAllocated = (UserAlloc.iAllocated + UserAlloc.iAllocIncrement);

	LeaveCriticalSection(&csMemLock);

	return FindFreeAllocSlot();
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CMemoryPool::Init(void)
{
	int iSlot = 0;

	InitializeCriticalSection(&csMemLock);
	
	memset(&UserAlloc, 0, sizeof(USERALLOC));
	memset(&UserAlloc.AllocPool, 0, sizeof(ALLOCPOOL));

	UserAlloc.iAllocated = 0;
	UserAlloc.iAllocIncrement = 100;

	if((UserAlloc.AllocPool = (ALLOCPOOL *) calloc(sizeof(ALLOCPOOL), UserAlloc.iAllocIncrement)) == NULL)
	{
		char sText[1024];
		sprintf(sText, "Error in Init: CMemoryPool Failed to initialize.");
		WriteAnEvent(sText, EVENT_ERROR);

		#ifdef _DEBUG_MEMPOOL
		__debugbreak();
		#endif

		return false;
	}

	UserAlloc.iAllocated = UserAlloc.iAllocIncrement;

	while(iSlot < UserAlloc.iAllocated)
	{
		UserAlloc.AllocPool[iSlot].bFree = true;
		UserAlloc.AllocPool[iSlot].iAddress = 0;
		iSlot++;
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CMemoryPool::Destroy(void)
{
	EnterCriticalSection(&csMemLock);

	FreeAll();

	free(UserAlloc.AllocPool);
		
	LeaveCriticalSection(&csMemLock);

	DeleteCriticalSection(&csMemLock);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CMemoryPool::Free(void *lpMemory)
{
	EnterCriticalSection(&csMemLock);

	int iSlot = 0;

	if(lpMemory == NULL)
	{
		char sText[1024];
		sprintf(sText, "Error in Free: Application attempted to free a NULL pointer.");
		WriteAnEvent(sText, EVENT_ERROR);

		#ifdef _DEBUG_MEMPOOL
		__debugbreak();
		#endif

		LeaveCriticalSection(&csMemLock);
		return false;
	}

	if((iSlot = GetSlotByAddress((int)lpMemory)) >= 0)
	{
		//Free the memory that the user allocated, also mark the AllocPool slot as free.
		free(lpMemory);
		
		lpMemory = NULL;

		UserAlloc.AllocPool[iSlot].bFree = true;
		UserAlloc.AllocPool[iSlot].iAddress = 0;
		iUserAllocs--;
	}
	else{
		char sText[1024];
		sprintf(sText, "Error in Free: Address [%d] not found in the managed memory pool.", lpMemory);
		WriteAnEvent(sText, EVENT_ERROR);

		#ifdef _DEBUG_MEMPOOL
		__debugbreak();
		#endif

		LeaveCriticalSection(&csMemLock);
		return false;
	}

	LeaveCriticalSection(&csMemLock);
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void *CMemoryPool::ReAllocate(void *lpMemory, const int iSize)
{
	EnterCriticalSection(&csMemLock);

	int iSlot = 0;
	char *sMemory = NULL;

	if((iSlot = GetSlotByAddress((int)lpMemory)) >= 0)
	{
		if((sMemory = (char *) realloc(lpMemory, iSize)) == NULL)
		{
			char sText[1024];
			sprintf(sText, "Error in ReAllocate: Failed to reallocate managed user memory for %d bytes.", iSize);
			WriteAnEvent(sText, EVENT_ERROR);

			#ifdef _DEBUG_MEMPOOL
			__debugbreak();
			#endif

			LeaveCriticalSection(&csMemLock);
			return NULL;
		}

		UserAlloc.AllocPool[iSlot].bFree = false;
		UserAlloc.AllocPool[iSlot].iAddress = (int) sMemory;
	}
	else{
		char sText[1024];
		sprintf(sText, "Error in ReAllocate: Address [%d] not found in the managed memory pool.", lpMemory);
		WriteAnEvent(sText, EVENT_ERROR);

		//Address not found in the memory pool.
		#ifdef _DEBUG_MEMPOOL
		__debugbreak();
		#endif

		LeaveCriticalSection(&csMemLock);
		return NULL;
	}

	LeaveCriticalSection(&csMemLock);

	return sMemory;	
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void *CMemoryPool::Allocate(const int iSize, const int iCount)
{
	EnterCriticalSection(&csMemLock);

	char *sMemory = NULL;
	int iSlot = 0;

	if((sMemory = (char *) calloc(iSize, iCount)) == NULL)
	{
		char sText[1024];
		sprintf(sText, "Error in Allocate: Failed to allocate managed user memory for %d bytes.", iSize * iCount);
		WriteAnEvent(sText, EVENT_ERROR);

		#ifdef _DEBUG_MEMPOOL
		__debugbreak();
		#endif

		LeaveCriticalSection(&csMemLock);
		return NULL;
	}

	iSlot = FindFreeAllocSlot();

	UserAlloc.AllocPool[iSlot].iAddress = (int) sMemory;
	UserAlloc.AllocPool[iSlot].bFree = false;

	iUserAllocs++;

	LeaveCriticalSection(&csMemLock);
	return sMemory;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#endif

